searchData={"items":[{"type":"module","title":"Eventful","doc":"This is the main Eventful module.\n\nEventful makes it easy to define your state machines on a given field on your schema.\n\nYou can even have multiple state machines on a single schema.\n\nLet's define the `Event` module.\n\n    defmodule MyApp.Post.Event do\n      alias MyApp.{\n        Post,\n        User\n      }\n\n      use Eventful,\n        parent: {:post, Post},\n        actor: {:user, User}\n\n      handle(:transitions, using: Post.Transitions)\n    end\n\nYou will need to define a `Transitions` module for `MyApp.Post`\n\n    defmodule MyApp.Post.Transitions do\n      use Eventful.Transition, repo: MyApp.Repo\n\n      @behaviour Eventful.Handler\n\n      alias MyApp.Post\n\n      Post\n      |> transition([from: \"created\", to: \"published\", via: \"publish\", fn changes ->\n        transit(changes)\n      end)\n\n      Post\n      |> transition([from: \"created\", to: \"deleted\", via: \"delete\", fn changes ->\n        transit(changes)\n      end)\n    end\n\nIn this example we're defining `Transitions` on `:current_state`. Events will be stored using the `Event` schema.\n\n    defmodule MyApp.Post do\n      use Ecto.Schema\n      import Ecto.Changeset\n\n      use Eventful.Transitable\n\n      alias __MODULE__.Event\n      alias __MODULE__.Transitions\n\n      Transitions\n      |> governs(:current_state, on: Event)\n\n      schema \"posts\" do\n        field :current_state, :string, default: \"created\"\n      end\n    end","ref":"Eventful.html"},{"type":"macro","title":"Eventful.handle/2","doc":"You can use the handle function to define the handler for a given domain:\n\n    defmodule MyApp.Post.Event do\n      alias MyApp.{\n        Post,\n        User\n      }\n\n      use Eventful,\n        parent: {:post, Post},\n        actor: {:user, User}\n\n      handle(:transitions, using: Post.Transitions)\n      handle(:visibilities, using: Post.Visibilities)\n    end","ref":"Eventful.html#handle/2"},{"type":"module","title":"Eventful.Error","doc":"","ref":"Eventful.Error.html"},{"type":"type","title":"Eventful.Error.t/0","doc":"","ref":"Eventful.Error.html#t:t/0"},{"type":"module","title":"Eventful.Exception","doc":"","ref":"Eventful.Exception.html"},{"type":"behaviour","title":"Eventful.Handler","doc":"","ref":"Eventful.Handler.html"},{"type":"callback","title":"Eventful.Handler.call/3","doc":"","ref":"Eventful.Handler.html#c:call/3"},{"type":"module","title":"Eventful.Metadata","doc":"This modules handles processing of the metadata for the event","ref":"Eventful.Metadata.html"},{"type":"function","title":"Eventful.Metadata.build/3","doc":"","ref":"Eventful.Metadata.html#build/3"},{"type":"protocol","title":"Eventful.Transit","doc":"","ref":"Eventful.Transit.html"},{"type":"function","title":"Eventful.Transit.perform/4","doc":"","ref":"Eventful.Transit.html#perform/4"},{"type":"type","title":"Eventful.Transit.t/0","doc":"All the types that implement this protocol.","ref":"Eventful.Transit.html#t:t/0"},{"type":"module","title":"Eventful.Transitable","doc":"This module is the one you'll use to define your main schema which require the state machine.\n\nFor example imagine the following:\n\n    defmodule MyApp.Post do\n      use Ecto.Schema\n      import Ecto.Changeset\n\n      schema \"posts\" do\n        field :current_state, :string, default: \"created\"\n      end\n    end\n\nYou can add a state machine like this assuming you've setup the `Event` module already:\n\n    defmodule MyApp.Post do\n      use Ecto.Schema\n      import Ecto.Changeset\n\n      use Eventful.Transitable\n\n      alias __MODULE__.Event\n      alias __MODULE__.Transitions\n\n      Transitions\n      |> governs(:current_state, on: Event)\n\n      schema \"posts\" do\n        field :current_state, :string, default: \"created\"\n      end\n    end\n\nYou can also setup multiple state machines with multiple fields:\n\n    defmodule MyApp.Post do\n      use Ecto.Schema\n      import Ecto.Changeset\n\n      use Eventful.Transitable\n\n      alias __MODULE__.Event\n      alias __MODULE__.Transitions\n      alias __MODULE__.Visibilities\n\n      # You can optionally use locks\n      Transitions\n      |> governs(:current_state, on: Event, lock: :current_state_version)\n\n      Visibilities\n      |> governs(:visibility, on: Event)\n\n      schema \"posts\" do\n        field :current_state, :string, default: \"created\"\n        field :current_state_version, :integer, default: 0\n\n        field :visibility, :string, default: \"private\"\n      end\n    end","ref":"Eventful.Transitable.html"},{"type":"macro","title":"Eventful.Transitable.governs/3","doc":"The governs function allows you to define governance on each of your fields.\n\n    Transitions\n    |> governs(:current_state, on: Event)","ref":"Eventful.Transitable.html#governs/3"},{"type":"module","title":"Eventful.Transition","doc":"This module providers the macros for building the transition modules\n\nHere is an example:\n\n    defmodule MyApp.Post.Transitions do\n      use Eventful.Transition, repo: MyApp.Repo\n\n      @behaviour Eventful.Handler\n\n      alias MyApp.Post\n\n      Post\n      |> transition([from: \"created\", to: \"published\", via: \"publish\", fn changes ->\n        transit(changes)\n      end)\n\n      Post\n      |> transition([from: \"created\", to: \"deleted\", via: \"delete\", fn changes ->\n        transit(changes)\n      end)\n    end\n\nYou can also define multiple state machines for example for a given `post` you can also have a machine for your `:visibility` field\n\n    defmodule MyApp.Post.Visibilities do\n      use Eventful.Transition, repo: MyApp.Repo, eventful_state: :visibility\n\n      @behaviour Eventful.Handler\n\n      alias MyApp.Post\n\n      Post\n      |> transition([from: \"private\", to: \"public\", via: \"publicize\", fn changes ->\n        transit(changes)\n      end)\n\n      Post\n      |> transition([from: \"public\", to: \"private\", via: \"privatize\", fn changes ->\n        transit(changes)\n      end)\n    end","ref":"Eventful.Transition.html"},{"type":"type","title":"Eventful.Transition.t/0","doc":"","ref":"Eventful.Transition.html#t:t/0"},{"type":"macro","title":"Eventful.Transition.transition/3","doc":"The transition macro allows you to define the state machine for your transitions\n\n    Post\n    |> transition([from: \"created\", to: \"published\", via: \"publish\", fn changes ->\n      transit(changes)\n    end)","ref":"Eventful.Transition.html#transition/3"},{"type":"module","title":"Eventful.Trigger","doc":"Handles the triggering of events\n\nYou can define a trigger in the following way\n\n    defmodule MyApp.Post.Triggers do\n      use Eventful.Trigger\n\n      alias MyApp.Post\n\n      Post\n      |> trigger([currently: \"published\"], fn event, post ->\n        # add your code here.\n      end)\n    end","ref":"Eventful.Trigger.html"},{"type":"macro","title":"Eventful.Trigger.trigger/3","doc":"The trigger macro function allows you to define a trigger in your trigger module\n\n    Post\n    |> trigger([currently: \"published\"], fn event, post ->\n      # add your code here.\n    end)","ref":"Eventful.Trigger.html#trigger/3"},{"type":"exception","title":"Eventful.Exception.UnexpectedReturnValue","doc":"","ref":"Eventful.Exception.UnexpectedReturnValue.html"},{"type":"extras","title":"Getting Started","doc":"# Getting Started\n\nEventful is a state machine library with an audit trail for your schemas. You can attach a state machine to any schema in your application.\n\nIn the following we will use a blogging app as an example. Let's imagine you had a schema like the following to store your blog post.\n\n```\ndefmodule MyApp.Post do\n  use Ecto.Schema\n  import Ecto.Changeset\n\n  schema \"posts\" do\n    field :title, :string\n    field :content, :string\n  end\n\n  def changeset(resource, attrs) do\n    resource\n    |> cast(attrs, [:title, :content])\n  end\nend\n```","ref":"getting-started.html"},{"type":"extras","title":"Event Schema - Getting Started","doc":"Let's imagine you want the ability to track the `state` of this post. You may have a collaboration feature where posts can be put into `draft` or `published` state, moreover you also want to track who did the transition. Let's assume you have a `User` schema of some kind. You could define an `Event` module like the following:\n\n```\ndefmodule MyApp.Post.UserEvent do\n  alias MyApp.{\n    Post,\n    User\n  }\n\n  use Eventful,\n    parent: {:post, Post},\n    actor: {:user, User},\n    table_name: \"post_user_events\"\nend\n```","ref":"getting-started.html#event-schema"},{"type":"extras","title":"Migration - Getting Started","doc":"To make this work you'll also need to add a migration.\n\n```\ndefmodule MyApp.Repo.Migrations.CreatePostUserEvents do\n  use Ecto.Migration\n\n  def change do\n    create table(:post_user_events) do\n      add(:name, :string, null: false)\n      add(:domain, :string, null: false)\n      add(:metadata, :map, default: \"{}\")\n\n      add(\n        :post_id,\n        references(:posts, on_delete: :restrict),\n        null: false\n      )\n\n      add(\n        :user_id,\n        references(:users, on_delete: :restrict),\n        null: false\n      )\n\n      timestamps()\n    end\n\n    create(index(:post_events, [:post_id]))\n    create(index(:post_events, [:user_id]))\n  end\nend\n```","ref":"getting-started.html#migration"},{"type":"extras","title":"State Machine - Getting Started","doc":"Next you'll need to define your `Transitions` this will allow you to define which states the post can transition to.\n\n```\ndefmodule MyApp.Post.Transitions do\n  use Eventful.Transition, repo: MyApp.Repo\n  \n  @behaviour Eventful.Handler\n  \n  alias MyApp.Post\n  \n  Post\n  |> transition([from: \"draft\", to: \"published\", via: \"publish\", fn changes ->\n    transit(changes)\n  end)\n  \n  Post\n  |> transition([from: \"published\", to: \"draft\", via: \"drafting\", fn changes ->\n    transit(changes)\n  end)\nend\n```\n\nNext you'll need to add some field to your `Post` schema which will be used to track the transitions. In this case let's add `:current_state` as the field and also define how the field is governed.\n\n```\ndefmodule MyApp.Post do\n  # ...\n  \n  use Eventful.Transitable\n  \n  alias __MODULE__.UserEvent\n  alias __MODULE__.Transitions\n  \n  Transitions\n  |> governs(:current_state, on: UserEvent)\n\n  schema \"posts\" do\n    field :title, :string\n    field :content, :string\n    \n    field :current_state, :string, default: \"draft\"\n  end\n\n  # ...\nend\n```\n\nAlso be sure to define the handler in your `UserEvent` module\n\n```\ndefmodule MyApp.Post.UserEvent do\n  alias MyApp.{\n    Post,\n    User\n  }\n\n  use Eventful,\n    parent: {:post, Post},\n    actor: {:user, User},\n    table_name: \"post_user_events\"\n    \n  handle(:transitions, using: Post.Transitions)\nend\n\ndefimpl Eventful.Transit, for: MyApp.Post do\n  alias MyApp.Post.UserEvent\n\n  def create(resource, actor, event_name, options \\\\ []) do\n    comment = Keyword.get(options, :comment)\n    parameters = Keyword.get(options, :parameters)\n\n    UserEvent.handle(post, user, %{\n      domain: \"transitions\", \n      event_name: event_name,\n      comment: comment,\n      parameters: parameters\n    })\n  end\nend\n```\n\nYou'll also need to add a migration for the post. You can use `:string` or if you prefer `:citext` for your `:current_state` field.\n\n```\ndefmodule MyApp.Repo.Migrations.AddCurrentStateToPosts do\n  use Ecto.Migration\n\n  def change do\n    alter table(:posts) do\n      add(:current_state, :citext, default: \"draft\", null: false)\n    end\n    \n    create(index(:posts, [:current_state]))\n  end\nend\n```","ref":"getting-started.html#state-machine"},{"type":"extras","title":"Transitioning the State - Getting Started","doc":"That's it! That's how your set up your first auditable state machine on your schema. You can how transition the post from state to state.\n\n```\n{:ok, transition} = Eventful.Transit.perform(post, user, \"publish\")\n```","ref":"getting-started.html#transitioning-the-state"},{"type":"extras","title":"Triggers","doc":"# Triggers\n\nTriggers are a powerful way to trigger some action that you define when a certain event occurs. Let's take a look at an example. Let's assume you have the following `Transitions` module.\n\n```\ndefmodule MyApp.Post.Transitions do\n  use Eventful.Transition, repo: MyApp.Repo\n  \n  @behaviour Eventful.Handler\n  \n  alias MyApp.Post\n  \n  Post\n  |> transition([from: \"draft\", to: \"published\", via: \"publish\", fn changes ->\n    transit(changes)\n  end)\n  \n  Post\n  |> transition([from: \"published\", to: \"draft\", via: \"drafting\", fn changes ->\n    transit(changes)\n  end)\nend\n```\n\nLet's say that when a post is `published` you would like your app to fire an email to notify someone that a post has been `published`.\n\nYou could define a trigger like this:\n\n```\ndefmodule MyApp.Post.Triggers do\n  alias MyApp.Post\n\n  use Eventful.Trigger\n\n  Post\n  |> trigger([currently: \"published\"], fn event, post ->\n    # user code to send email\n  end)\nend\n```\n\nYou will also need to update your `Transitions` module like so:\n\n```\ndefmodule MyApp.Post.Transitions do\n  use Eventful.Transition, repo: MyApp.Repo\n  \n  @behaviour Eventful.Handler\n  \n  alias MyApp.Post\n  \n  Post\n  |> transition([from: \"draft\", to: \"published\", via: \"publish\", fn changes ->\n    transit(changes, Post.Triggers)\n  end)\n  \n  Post\n  |> transition([from: \"published\", to: \"draft\", via: \"drafting\", fn changes ->\n    transit(changes)\n  end)\nend\n```\n\nNow every time a post is published it will send the email notification.","ref":"triggers.html"},{"type":"extras","title":"Guards","doc":"# Guards\n\nGuards are useful when you wish to prevent your transition from happening based on some kind of check.\n\nIn your `Transitions` module you can define guards like so:\n\n```\ndefmodule MyApp.Post.Transitions do\n  alias MyApp.Post\n\n  @behaviour Eventful.Handler\n\n  use Eventful.Transition, repo: MyApp.Repo\n\n  Post\n  |> transition(\n    [from: \"draft\", to: \"published\", via: \"publish\"],\n    fn {post_changeset, _} = changes ->\n      if MyApp.check_something(post_changeset.data) do\n        transit(changes, Post.Triggers)\n      else\n        {:error, \"guard failed\"}\n      end\n    end)\n  )\nend\n```\n\nThis allows you to define that a given condition computed by `MyApp.check_something/1` will need to return true when `publish` event is invoked.","ref":"guards.html"}],"proglang":"elixir","content_type":"text/markdown","producer":{"name":"ex_doc","version":"0.38.1"}}